package ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures;

import ninja.thefirearchmage.games.fourtykcalculator.utils.functions.Function4;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.ToString;

import java.util.function.Function;

/**
 * Data structure to hold 4 generic fields.
 * @param <T> First field type
 * @param <R> Second field type
 * @param <U> Third field type
 * @param <Y> Fourth field type
 */
@Getter
@ToString(of = {"_1", "_2", "_3", "_4"})
@EqualsAndHashCode(of = {"_1", "_2", "_3", "_4"}, callSuper = false)
public class Tuple4<T,R,U,Y> extends Tuple {
    final public T _1;
    final public R _2;
    final public U _3;
    final public Y _4;

    Tuple4(T _1, R _2, U _3, Y _4) {
        this._1 = _1;
        this._2 = _2;
        this._3 = _3;
        this._4 = _4;
    }

    /**
     * Maps the fields of the tuple with the given functions. Does not handle mapping errors.
     * @return a mapped object from the original tuple
     */
    public <A> A map(Function4<T, R, U,Y, A> tuppleMapper) {
        return tuppleMapper.apply(_1, _2, _3, _4);
    }

    /**
     * Returns a new tuple with each field being the result of applying the corresponding mapper function.
     * @param f1Mapper the mapping function to apply to the first field
     * @param f2Mapper the mapping function to apply to the second field
     * @param f3Mapper the mapping function to apply to the third field
     * @param f4Mapper the mapping function to apply to the fourth field
     * @return a new tuple with the mapper functions applied
     * @param <A> the new return type for the first field
     * @param <S> the new return type for the second field
     * @param <D> the new return type for the third field
     * @param <F> the new return type for the fourth field
     */
    public <A,S,D,F> Tuple4<A,S,D,F> map(Function<T,A> f1Mapper,
                                         Function<R,S> f2Mapper,
                                         Function<U,D> f3Mapper,
                                         Function<Y,F> f4Mapper) {
        return Tuple.of(f1Mapper.apply(_1), f2Mapper.apply(_2), f3Mapper.apply(_3), f4Mapper.apply(_4));
    }

    public Tuple3<R,U,Y> remove1() {
        return Tuple.of(_2, _3, _4);
    }
    public Tuple3<T,U,Y> remove2() {
        return Tuple.of(_1, _3, _4);
    }
    public Tuple3<T,R,Y> remove3() {
        return Tuple.of(_1, _2, _4);
    }
    public Tuple3<T,R,U> remove4() {
        return Tuple.of(_1, _2, _3);
    }
}
